// mission controller node all-in-one for DJI M210 (v1 or v2)

#include <mission_controller/mission_controller_node.h>

// #include <dji_osdk_ros/vehicle_wrapper.h>
// #include <dji_osdk_ros/dji_vehicle_node.h>

// global variables
ros::ServiceClient     waypoint_upload_client;
ros::ServiceClient     waypoint_action_client;
ros::ServiceClient     flight_control_client;
ros::ServiceClient     mission_waypoint_client;
ros::ServiceClient     obtain_ctrl_authority_client;
// need to take another look at synchronization...
// ros::Publisher         syncPub; // thing to publish to sync topic "positionSync"

sensor_msgs::NavSatFix gps_pos; // gps position from drone
geometry_msgs::Transform ex_pos; // position from external source (qualisys or some such)
geometry_msgs::Transform track_pos; // target position
geometry_msgs::Transform previous_ex_pos;
geometry_msgs::Transform previous_track_pos;

geometry_msgs::Transform pd_pos; // desired pose from pd relative to qual center (only x on rotation)

std::vector<float> expectedVel; // expected position after applying relative position move
float feedbackYaw; // yaw as measured by pos ref
float feedbackPitch; // pitch as measured by pos ref
float feedbackRoll; // roll as measured by pos ref
float trackingYaw; // yaw to track
float trackingPitch; // pitch to track
float trackingRoll; // roll to track
int calibrationWindow = 3; // the last _ position calls that are included in coordinate system calibration
std::vector<float> calibrationWindowValues;
float angleOffset;
float controlLoopTime;
float updateRateParam;
std::vector<std::vector<float>> poseRotationMatrix; // need to run calibration to get
std::vector<std::vector<std::vector<float>>> poseRotationMatrixValues; // need to run calibration to get
float poseMagnitudeScaler; // need to run calibration to get
float poseRotationRadians;
bool rotationMatrixCalibratedFlag = false;

bool                   gps_home_flag = false;
bool                   pos_home_flag = false;
bool                   localMissionFlag = false; // whether running position mission currently
bool                   gpsMissionFlag = false; // whether running gps mission currently
std::string                 missionTypeParam;
ros::Subscriber        gps_pos_subscriber;
ros::Subscriber        pos_feedback_subscriber;
ros::Subscriber        pos_tracking_subscriber;
ros::Subscriber        pd_position_subscriber;
ros::Subscriber        pd_heading_subscriber;

// temp subscribers to pick up puredata parameter information, will be removed once the right parameters are settled on
ros::Subscriber        pd_parameter_MinSpeed_subscriber;
ros::Subscriber        pd_parameter_MaxSpeed_subscriber;
ros::Subscriber        pd_parameter_MaxYawSpeed_subscriber;
ros::Subscriber        pd_parameter_SpeedScale_subscriber;
ros::Subscriber        pd_parameter_UpdateRate_subscriber;
ros::Subscriber        pd_parameter_Damping_subscriber;
ros::Subscriber        pd_parameter_Frequency_subscriber;
ros::Subscriber        pd_parameter_AngleOffset_subscriber;

int                    droneId; // stupid but will work for now

int                    waypointMissionUpdateCounter = 0; // counts iterations before we check waypoints again
int                    waypointMissionUpdateLimit = 10; // number of counts before we check ...
int                    positionMissionUpdateCounter = 0; // counts iterations before we check positions again
int                    positionMissionUpdateLimit = 10; // ...

                                    // std::vector<WayPointSettings> generatedWaypts; // waypoints to be generated, used to check if we're at waypoint or not
int                    waypointNumber;
int                    currentWaypoint = 0; // how many waypoints have we reached
long double            initialWaypointDistance; // distance to first waypoint
long double            initialPositionDistance; // distance to first position
double                 waypointSpeedParam;
double                 positionThresholdParam;
double                 angleOffsetParam;
double		             maxSpeedParam;
double                 speedScalarParam;
double                 maxYawSpeedParam;

// inverse controller variables
double                 dt;
double                 damping;
double                 frequency;
double                 kp;
double                 kd;
double                 ksg;
double                 kdg;
double                 g;


int main(int argc, char** argv)
{

  ros::init(argc, argv, "mission_controller");
  ros::NodeHandle nh;
  //
  // ros::param::get("base_speed", waypointSpeedParam);
  // ros::param::get("max_speed", maxSpeedParam);
  // ros::param::get("max_yaw_speed", maxYawSpeedParam);
  // ros::param::get("angle_offset", angleOffsetParam);
  // ros::param::get("speed_scalar", speedScalarParam);
  // ros::param::get("drone_id", droneId);
  // ros::param::get("mission_type", missionTypeParam);
  // ros::param::get("position_threshold", positionThresholdParam);
  // ros::param::get("loop_rate", updateRateParam);
  // controlLoopTime = 1.0/updateRateParam;
  // // additions
  // ros::param::get("controller_damping", damping);
  // ros::param::get("controller_frequency", frequency);
  // kp = (6.0*frequency)*(6.0*frequency)* 0.25;
  // kd = 4.5*frequency*damping;
  // expectedVel = std::vector<float> {0.0, 0.0, 0.0};
  // angleOffset = 0.0;
  // calibrationWindowValues = std::vector<float> {};
  // for (int i = 0; i < calibrationWindow; i++)
  // {
  //   calibrationWindowValues.push_back(0.0);
  // }
  //
  // // set puredata position to zero zero so it doesn't immediately fly away
  // pd_pos.translation.x = 0.0;
  // pd_pos.translation.y = 0.0;
  // pd_pos.translation.z = 1.5;
  // pd_pos.rotation.x = 0.0;
  //
  // // ROS stuff
  // // obtain control authority?
  // obtain_ctrl_authority_client = nh.serviceClient<dji_osdk_ros::ObtainControlAuthority>(
  //   "obtain_release_control_authority");
  // waypoint_upload_client = nh.serviceClient<dji_osdk_ros::MissionWpUpload>(
  //   "dji_osdk_ros/mission_waypoint_upload");
  // waypoint_action_client = nh.serviceClient<dji_osdk_ros::MissionWpAction>(
  //   "dji_osdk_ros/mission_waypoint_action");
  // flight_control_client = nh.serviceClient<dji_osdk_ros::FlightTaskControl>(
  //   "flight_task_control");
  //
  // // syncPub = nh.advertise<std_msgs::Float32>("sync_signal", 1);
  // gps_pos_subscriber = nh.subscribe<sensor_msgs::NavSatFix>("dji_osdk_ros/gps_position", 1, &gpsPosCallback);
  // pos_feedback_subscriber = nh.subscribe<geometry_msgs::Transform>("feedback", 1, &posFeedbackCallback);
  // pos_tracking_subscriber = nh.subscribe<geometry_msgs::Transform>("tracking", 1, &posTrackingCallback);
  // pd_position_subscriber = nh.subscribe<geometry_msgs::Transform>("pd/target", 1, &pdPositionCallback);
  //
  // // pd_parameter_MinSpeed_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/minSpeed", 1, &pdMinSpeedCallback);
  // pd_parameter_MaxSpeed_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/maxSpeed", 1, &pdMaxSpeedCallback);
  // pd_parameter_MaxYawSpeed_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/maxYawSpeed", 1, &pdMaxYawSpeedCallback);
  // pd_parameter_SpeedScale_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/speedScale", 1, &pdSpeedScaleCallback);
  // pd_parameter_UpdateRate_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/updateRate", 1, &pdUpdateRateCallback);
  // pd_parameter_AngleOffset_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/angleOffset", 1, &pdAngleOffsetCallback);
  // pd_parameter_Damping_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/controllerDamping", 1, &pdDampingCallback);
  // pd_parameter_Frequency_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/controllerFrequency", 1, &pdFrequencyCallback);
  //
  // ros::ServiceServer overwatchService = nh.advertiseService("overwatch", overwatchFunction);
  //
  // // wait for services to be up
  // obtain_ctrl_authority_client.waitForExistence();
  // waypoint_upload_client.waitForExistence();
  // waypoint_action_client.waitForExistence();
  // flight_control_client.waitForExistence();
  // // wait for waypoints to be up as well
  ROS_INFO("mission controller ready to go");

  ros::spin();

  return 0;
}
